/* eslint no-redeclare: "off" */

var utils = require('../utils');

function IDirectory (sc, path, handle, fs) {
	this.sc = sc;
	this.fs = fs;
	this.handle = handle;
	this.path = path;
}

IDirectory.prototype.GetEntryCount = function () {
	return this.sc.ipcMsg(1).sendTo(this.handle).asResult().map((r) => [r.dataBuffer[0], r.dataBuffer[1]]);
};

IDirectory.prototype.GetEntries = function (buf, numEntries) {
	if(buf.byteLength < 0x310 * numEntries) {
		throw new Error("buffer too small");
	}
	return this.sc.ipcMsg(0).data(0).bDescriptor(buf, 0x310 * numEntries, 0).sendTo(this.handle).asResult();
};

IDirectory.prototype.DirList = function (indentation) {
	var s = '';

	if (indentation !== undefined) {
		for (var i = 0; i < indentation; i++) {
			s += '   ';
		}
	}

	var entryCount = utils.trunc32(this.GetEntryCount().assertOk());

	if (entryCount > 0) {
		var entryBuf = new Uint32Array(0x310 * entryCount);
		this.GetEntries(entryBuf, entryCount).assertOk();
		for (i = 0; i < entryCount; i++) {
			var fn = this.path + utils.u8a2str(new Uint8Array(entryBuf.buffer, 0x310 * i, 0x300));
			var eType = entryBuf[(0x310 * i + 0x304) >> 2];
			if (eType === 1) {
				utils.log(s + '<FILE> ' + fn);
			} else {
				utils.log(s + '<FOLDER> ' + fn + '/');
				var f = this.fs.OpenDir(fn + '/').assertOk();
				try {
					f.DirList(indentation + 1);
				} finally {
					f.Close();
				}
			}
		}
	}
};

IDirectory.prototype.DirDump = function (dumpPath) {
	var entryCount = utils.trunc32(this.GetEntryCount().assertOk());
	if (entryCount > 0) {
		var entryBuf = new Uint32Array(0x310 * entryCount);
		this.GetEntries(entryBuf, entryCount).assertOk();
		for (var i = 0; i < entryCount; i++) {
			var fn = this.path + utils.u8a2nullstr(new Uint8Array(entryBuf.buffer, 0x310 * i, 0x300));
			var eType = entryBuf[(0x310 * i + 0x304) >> 2];
			if (eType === 1) {
				utils.log('<FILE> ' + fn);
				var fp = this.fs.OpenFile(fn).assertOk();
				try {
					var buf = fp.ReadAll().assertOk();
					this.sc.memdump(buf, fp.GetSize().assertOk(), dumpPath + fn);
				} finally {
					fp.Close();
				}
			} else {
				utils.log('<FOLDER> ' + fn + '/');
				var f = this.fs.OpenDir(fn + '/').assertOk();
				try {
					f.DirDump(dumpPath);
				} finally {
					f.Close();
				}
			}
		}
	}
};

IDirectory.prototype.Close = function () {
	return this.sc.svcCloseHandle(this.handle);
};

module.exports = IDirectory;
