var utils = require('../utils');

function IFile (sc, handle) {
	this.sc = sc;
	this.handle = handle;
}

IFile.prototype.Read = function (buffer, offset, length) {
	if(offset === undefined) {
		offset = [0, 0];
	}
	if(length === undefined) {
		length = buffer.byteLength;
	}
	return this.sc.ipcMsg(0).datau64(0, offset, length).bDescriptor(buffer, length, 1).sendTo(this.handle).asResult().replaceValue(buffer);
};

IFile.prototype.Write = function (offset, buf, size) {
	return this.sc.ipcMsg(1).aDescriptor(buf, size, 1).datau64(0, offset, size).sendTo(this.handle).asResult();
};

IFile.prototype.Flush = function() {
	return this.sc.ipcMsg(2).sendTo(this.handle).asResult();
};

IFile.prototype.SetSize = function (size) {
	return this.sc.ipcMsg(3).datau64(size).sendTo(this.handle).asResult();
};

IFile.prototype.GetSize = function () {
	return this.sc.ipcMsg(4).sendTo(this.handle).asResult()
		.map((r) => [r.data[0], r.data[1]]);
};

IFile.prototype.ReadAll = function () {
	var self = this;
	return this.GetSize().andThen((size) => {
		var fSize = utils.trunc32(size);
		var m = new ArrayBuffer(fSize);
		return self.sc.ipcMsg(0).datau64(0, 0, fSize).bDescriptor(m, fSize, 1).sendTo(self.handle).asResult().replaceValue(m);
	});
};

IFile.prototype.Close = function () {
	return this.sc.svcCloseHandle(this.handle);
};

// for SploitCore.prototype.dumpToFile
IFile.prototype.makeReader = function() {
	var f = this;
	return (ab, offset, size) => {
		return f.Read(ab, offset, size);
	};
};

module.exports = IFile;
